from unittest import TestCase

import os
import sys
from mock import MagicMock
from mock import patch
from autosubmit.database.db_manager import DbManager


class TestDbManager(TestCase):
    def test_create_table_command_returns_a_valid_command(self):
        # arrange
        table_name = 'tests'
        table_fields = ['dummy1', 'dummy2', 'dummy3']
        expected_command = 'CREATE TABLE IF NOT EXISTS tests (dummy1, dummy2, dummy3)'
        # act
        command = DbManager.generate_create_table_command(table_name, table_fields)
        # assert
        self.assertEqual(expected_command, command)

    def test_insert_command_returns_a_valid_command(self):
        # arrange
        table_name = 'tests'
        columns = ['col1, col2, col3']
        values = ['dummy1', 'dummy2', 'dummy3']
        expected_command = 'INSERT INTO tests(col1, col2, col3) VALUES ("dummy1", "dummy2", "dummy3")'
        # act
        command = DbManager.generate_insert_command(table_name, columns, values)
        # assert
        self.assertEqual(expected_command, command)

    def test_insert_many_command_returns_a_valid_command(self):
        # arrange
        table_name = 'tests'
        num_of_values = 3
        expected_command = 'INSERT INTO tests VALUES (?,?,?)'
        # act
        command = DbManager.generate_insert_many_command(table_name, num_of_values)
        # assert
        self.assertEqual(expected_command, command)

    def test_select_command_returns_a_valid_command(self):
        # arrange
        table_name = 'tests'
        where = ['test=True', 'debug=True']
        expected_command = 'SELECT * FROM tests WHERE test=True AND debug=True'
        # act
        command = DbManager.generate_select_command(table_name, where)
        # assert
        self.assertEqual(expected_command, command)

    def test_when_database_already_exists_then_is_not_initialized_again(self):
        sys.modules['os'].path.exists = MagicMock(return_value=True)
        connection_mock = MagicMock()
        cursor_mock = MagicMock()
        cursor_mock.side_effect = Exception('This method should not be called')
        connection_mock.cursor = MagicMock(return_value=cursor_mock)
        original_connect = sys.modules['sqlite3'].connect
        sys.modules['sqlite3'].connect = MagicMock(return_value=connection_mock)
        DbManager('dummy-path', 'dummy-name', 999)
        connection_mock.cursor.assert_not_called()
        sys.modules['sqlite3'].connect = original_connect



#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS
# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import collections

JobDataRow = collections.namedtuple('JobDataRow', ['id', 'counter', 'job_name', 'created', 'modified', 'submit', 'start', 'finish', 
                                                  'status', 'rowtype', 'ncpus', 'wallclock', 'qos', 'energy', 'date', 'section', 'member', 
                                                  'chunk', 'last', 'platform', 'job_id', 'extra_data', 'nnodes', 'run_id', 'MaxRSS', 'AveRSS', 
                                                  'out', 'err', 'rowstatus', 'children', 'platform_output'])

ExperimentRunRow = collections.namedtuple('ExperimentRunRow', [
                                           'run_id', 'created', 'modified', 'start', 'finish', 'chunk_unit', 'chunk_size', 'completed', 'total', 'failed', 'queuing', 'running', 'submitted', 'suspended', 'metadata'])

ExperimentStatusRow = collections.namedtuple(
    'ExperimentStatusRow', ['exp_id', 'name', 'status', 'seconds_diff', 'modified'])

ExperimentRow = collections.namedtuple('ExperimentRow', ["id", "name", "autosubmit_version", "description"])

PragmaVersion = collections.namedtuple('PragmaVersion', ['version'])
MaxCounterRow = collections.namedtuple('MaxCounter', ['maxcounter'])

class RunningStatus:
  RUNNING = "RUNNING"
  NOT_RUNNING = "NOT RUNNING"

class RowType:
    NORMAL = 2
    # PACKED = 2

class RowStatus:
    INITIAL = 0
    COMPLETED = 1    
    PROCESSED = 2
    FAULTY = 3
    CHANGED = 4
    PENDING_PROCESS = 5

table_name_to_model = {
  "experiment" : ExperimentRow,
  "experiment_status" : ExperimentStatusRow,
  "job_data" : JobDataRow,
  "experiment_run" : ExperimentRunRow,
  "pragma_version" : PragmaVersion
}




#!/usr/bin/python

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS
# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import traceback
import os
import time
from shutil import copy2
from collections import namedtuple
from autosubmit.history.experiment_history import ExperimentHistory
from autosubmit.history.internal_logging import Logging
from autosubmit.history.strategies import StraightWrapperAssociationStrategy, GeneralizedWrapperDistributionStrategy, PlatformInformationHandler
from autosubmitconfigparser.config.basicconfig import BasicConfig
from autosubmit.history.platform_monitor.slurm_monitor import SlurmMonitor
EXPID_TT00_SOURCE = "test_database.db~"
EXPID_TT01_SOURCE = "test_database_no_run.db~"
EXPID = "tt00"
EXPID_NONE = "tt01"
BasicConfig.read()
JOBDATA_DIR = BasicConfig.JOBDATA_DIR
LOCAL_ROOT_DIR = BasicConfig.LOCAL_ROOT_DIR
job = namedtuple("Job", ["name", "date", "member", "status_str", "children"])

@unittest.skip('TODO: another test that uses actual data. See if there is anything useful, and extract into functional/integration/unit tests that run on any machine')
class TestExperimentHistory(unittest.TestCase):
  # @classmethod
  # def setUpClass(cls):    
  #   cls.exp = ExperimentHistory("tt00") # example database
  def setUp(self):
    source_path_tt00 = os.path.join(JOBDATA_DIR, EXPID_TT00_SOURCE)
    self.target_path_tt00 = os.path.join(JOBDATA_DIR, "job_data_{0}.db".format(EXPID))        
    copy2(source_path_tt00, self.target_path_tt00)
    source_path_tt01 = os.path.join(JOBDATA_DIR, EXPID_TT01_SOURCE)
    self.target_path_tt01 = os.path.join(JOBDATA_DIR, "job_data_{0}.db".format(EXPID_NONE))
    copy2(source_path_tt01, self.target_path_tt01)
    self.job_list = [
      job("a29z_20000101_fc2_1_POST", "2000-01-01 00:00:00", "POST", "COMPLETED", ""),
      job("a29z_20000101_fc1_1_CLEAN", "2000-01-01 00:00:00", "CLEAN", "COMPLETED", ""),
      job("a29z_20000101_fc3_1_POST", "2000-01-01 00:00:00", "POST", "RUNNING", ""),
      job("a29z_20000101_fc2_1_CLEAN", "2000-01-01 00:00:00", "CLEAN", "COMPLETED", ""),
      job("a29z_20000101_fc0_3_SIM", "2000-01-01 00:00:00", "SIM", "COMPLETED", ""),
      job("a29z_20000101_fc1_2_POST", "2000-01-01 00:00:00", "POST", "QUEUING", ""),
    ] # 2 differences, all COMPLETED
    self.job_list_large = [
      job("a29z_20000101_fc2_1_POST", "2000-01-01 00:00:00", "POST", "COMPLETED", ""),
      job("a29z_20000101_fc1_1_CLEAN", "2000-01-01 00:00:00", "CLEAN", "COMPLETED", ""),
      job("a29z_20000101_fc3_1_POST", "2000-01-01 00:00:00", "POST", "RUNNING", ""),
      job("a29z_20000101_fc2_1_CLEAN", "2000-01-01 00:00:00", "CLEAN", "COMPLETED", ""),
      job("a29z_20000101_fc0_3_SIM", "2000-01-01 00:00:00", "SIM", "COMPLETED", ""),
      job("a29z_20000101_fc1_2_POST", "2000-01-01 00:00:00", "POST", "QUEUING", ""),
      job("a29z_20000101_fc1_5_POST", "2000-01-01 00:00:00", "POST", "SUSPENDED", ""),
      job("a29z_20000101_fc1_4_POST", "2000-01-01 00:00:00", "POST", "FAILED", ""),
      job("a29z_20000101_fc2_5_CLEAN", "2000-01-01 00:00:00", "CLEAN", "SUBMITTED", ""),
      job("a29z_20000101_fc0_1_POST", "2000-01-01 00:00:00", "POST", "RUNNING", ""),
    ]
  
  def tearDown(self):
    os.remove(self.target_path_tt00)
    os.remove(self.target_path_tt01)

  def test_db_exists(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    self.assertTrue(exp_history.manager.my_database_exists() is True)
    exp_history = ExperimentHistory("tt99")
    self.assertTrue(exp_history.manager.my_database_exists() is False)
  
  def test_is_header_ready(self):
    exp_history = ExperimentHistory("tt00")
    self.assertTrue(exp_history.is_header_ready() is True)

  def test_detect_differences_job_list(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    differences = exp_history.detect_changes_in_job_list(self.job_list)
    expected_differences = ["a29z_20000101_fc3_1_POST", "a29z_20000101_fc1_2_POST"]
    for job_dc in differences:
      self.assertTrue(job_dc.job_name in expected_differences)
    self.assertTrue(len(differences) == 2)

  def test_built_list_of_changes(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    built_differences = exp_history._get_built_list_of_changes(self.job_list)
    expected_ids_differences = [90, 101]
    for item in built_differences:
      self.assertTrue(item[3] in expected_ids_differences)
  
  def test_get_date_member_count(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    dm_count = exp_history._get_date_member_completed_count(self.job_list)    
    self.assertTrue(dm_count > 0)

  def test_should_we_create_new_run(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    CHANGES_COUNT = 1
    TOTAL_COUNT = 6
    current_experiment_run_dc = exp_history.manager.get_experiment_run_dc_with_max_id()    
    current_experiment_run_dc.total = TOTAL_COUNT
    should_we = exp_history.should_we_create_a_new_run(self.job_list, CHANGES_COUNT, current_experiment_run_dc, current_experiment_run_dc.chunk_unit, current_experiment_run_dc.chunk_size)
    self.assertTrue(should_we is False)
    TOTAL_COUNT_DIFF = 5
    current_experiment_run_dc.total = TOTAL_COUNT_DIFF
    should_we = exp_history.should_we_create_a_new_run(self.job_list, CHANGES_COUNT, current_experiment_run_dc, current_experiment_run_dc.chunk_unit, current_experiment_run_dc.chunk_size)
    self.assertTrue(should_we is True)
    CHANGES_COUNT = 5
    should_we = exp_history.should_we_create_a_new_run(self.job_list, CHANGES_COUNT, current_experiment_run_dc, current_experiment_run_dc.chunk_unit, current_experiment_run_dc.chunk_size)
    self.assertTrue(should_we is True)
    CHANGES_COUNT = 1
    current_experiment_run_dc.total = TOTAL_COUNT
    should_we = exp_history.should_we_create_a_new_run(self.job_list, CHANGES_COUNT, current_experiment_run_dc, current_experiment_run_dc.chunk_unit, current_experiment_run_dc.chunk_size*20)
    self.assertTrue(should_we is True)
    should_we = exp_history.should_we_create_a_new_run(self.job_list, CHANGES_COUNT, current_experiment_run_dc, current_experiment_run_dc.chunk_unit, current_experiment_run_dc.chunk_size)
    self.assertTrue(should_we is False)
    should_we = exp_history.should_we_create_a_new_run(self.job_list, CHANGES_COUNT, current_experiment_run_dc, "day", current_experiment_run_dc.chunk_size)
    self.assertTrue(should_we is True)

  def test_status_counts(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    result = exp_history.get_status_counts_from_job_list(self.job_list_large)
    self.assertTrue(result["COMPLETED"] == 4)
    self.assertTrue(result["QUEUING"] == 1)
    self.assertTrue(result["RUNNING"] == 2)
    self.assertTrue(result["FAILED"] == 1)
    self.assertTrue(result["SUSPENDED"] == 1)
    self.assertTrue(result["TOTAL"] == len(self.job_list_large))
  
  def test_create_new_experiment_run_with_counts(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    exp_run = exp_history.create_new_experiment_run(job_list=self.job_list)
    self.assertTrue(exp_run.chunk_size == 0)
    self.assertTrue(exp_run.chunk_unit == "NA")
    self.assertTrue(exp_run.total == len(self.job_list))
    self.assertTrue(exp_run.completed == 4)
  
  def test_finish_current_run(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    exp_run = exp_history.finish_current_experiment_run()    
    self.assertTrue(len(exp_run.modified) > 0)    
    self.assertTrue(exp_run.finish > 0)

  def test_process_job_list_changes(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    exp_run = exp_history.process_job_list_changes_to_experiment_totals(self.job_list)
    self.assertTrue(exp_run.total == len(self.job_list))
    self.assertTrue(exp_run.completed == 4)
    self.assertTrue(exp_run.running == 1)
    self.assertTrue(exp_run.queuing == 1)

  def test_calculated_weights(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    job_data_dcs = exp_history.manager.get_all_last_job_data_dcs()    
    calculated_weights = GeneralizedWrapperDistributionStrategy().get_calculated_weights_of_jobs_in_wrapper(job_data_dcs)
    sum_comp_weight = 0
    for job_name in calculated_weights:      
      sum_comp_weight += calculated_weights[job_name]    
    self.assertTrue(abs(sum_comp_weight - 1) <= 0.01)
  
  def test_distribute_energy_in_wrapper_1_to_1(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    ssh_output = '''                 17857525  COMPLETED         10        1 2021-10-13T15:51:16 2021-10-13T15:51:17 2021-10-13T15:52:47         2.41K                                                     
           17857525.batch  COMPLETED         10        1 2021-10-13T15:51:17 2021-10-13T15:51:17 2021-10-13T15:52:47          1.88K                     6264K                     6264K 
          17857525.extern  COMPLETED         10        1 2021-10-13T15:51:17 2021-10-13T15:51:17 2021-10-13T15:52:47          1.66K                      473K                       68K 
               17857525.0  COMPLETED         10        1 2021-10-13T15:51:21 2021-10-13T15:51:21 2021-10-13T15:51:22            186                      352K                   312.30K 
               17857525.1  COMPLETED         10        1 2021-10-13T15:51:23 2021-10-13T15:51:23 2021-10-13T15:51:24            186                      420K                   306.70K 
               17857525.2  COMPLETED         10        1 2021-10-13T15:51:24 2021-10-13T15:51:24 2021-10-13T15:51:27            188                      352K                   325.80K 
               17857525.3  COMPLETED         10        1 2021-10-13T15:51:28 2021-10-13T15:51:28 2021-10-13T15:51:29            192                      352K                   341.90K                
    '''
    slurm_monitor = SlurmMonitor(ssh_output)
    job_data_dcs = exp_history.manager.get_all_last_job_data_dcs()[:4] # Get me 4 jobs
    weights = StraightWrapperAssociationStrategy().get_calculated_weights_of_jobs_in_wrapper(job_data_dcs)    
    info_handler = PlatformInformationHandler(StraightWrapperAssociationStrategy())
    job_data_dcs_with_data = info_handler.execute_distribution(job_data_dcs[0], job_data_dcs, slurm_monitor)    
    self.assertTrue(job_data_dcs_with_data[0].energy == round(slurm_monitor.steps[0].energy + weights[job_data_dcs_with_data[0].job_name]*slurm_monitor.extern.energy, 2))
    self.assertTrue(job_data_dcs_with_data[0].MaxRSS == slurm_monitor.steps[0].MaxRSS)
    self.assertTrue(job_data_dcs_with_data[2].energy == round(slurm_monitor.steps[2].energy + weights[job_data_dcs_with_data[2].job_name]*slurm_monitor.extern.energy, 2))
    self.assertTrue(job_data_dcs_with_data[2].AveRSS == slurm_monitor.steps[2].AveRSS)

  def test_distribute_energy_in_wrapper_general_case(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    ssh_output = '''                 17857525  COMPLETED         10        1 2021-10-13T15:51:16 2021-10-13T15:51:17 2021-10-13T15:52:47         2.41K                                                     
           17857525.batch  COMPLETED         10        1 2021-10-13T15:51:17 2021-10-13T15:51:17 2021-10-13T15:52:47          1.88K                     6264K                     6264K 
          17857525.extern  COMPLETED         10        1 2021-10-13T15:51:17 2021-10-13T15:51:17 2021-10-13T15:52:47          1.66K                      473K                       68K 
               17857525.0  COMPLETED         10        1 2021-10-13T15:51:21 2021-10-13T15:51:21 2021-10-13T15:51:22            186                      352K                   312.30K 
               17857525.1  COMPLETED         10        1 2021-10-13T15:51:23 2021-10-13T15:51:23 2021-10-13T15:51:24            186                      420K                   306.70K 
               17857525.2  COMPLETED         10        1 2021-10-13T15:51:24 2021-10-13T15:51:24 2021-10-13T15:51:27            188                      352K                   325.80K 
               17857525.3  COMPLETED         10        1 2021-10-13T15:51:28 2021-10-13T15:51:28 2021-10-13T15:51:29            192                      352K                   341.90K                
    '''
    slurm_monitor = SlurmMonitor(ssh_output)
    job_data_dcs = exp_history.manager.get_all_last_job_data_dcs()[:5] # Get me 5 jobs
    weights = GeneralizedWrapperDistributionStrategy().get_calculated_weights_of_jobs_in_wrapper(job_data_dcs)    
    # print(sum(weights[k] for k in weights))
    info_handler = PlatformInformationHandler(GeneralizedWrapperDistributionStrategy())
    job_data_dcs_with_data = info_handler.execute_distribution(job_data_dcs[0], job_data_dcs, slurm_monitor)
    self.assertTrue(job_data_dcs_with_data[0].energy == round(slurm_monitor.total_energy * weights[job_data_dcs_with_data[0].job_name], 2))
    self.assertTrue(job_data_dcs_with_data[1].energy == round(slurm_monitor.total_energy * weights[job_data_dcs_with_data[1].job_name], 2))
    self.assertTrue(job_data_dcs_with_data[2].energy == round(slurm_monitor.total_energy * weights[job_data_dcs_with_data[2].job_name], 2))
    self.assertTrue(job_data_dcs_with_data[3].energy == round(slurm_monitor.total_energy * weights[job_data_dcs_with_data[3].job_name], 2))
    self.assertTrue(job_data_dcs_with_data[4].energy == round(slurm_monitor.total_energy * weights[job_data_dcs_with_data[4].job_name], 2))
    sum_energy = sum(job.energy for job in job_data_dcs_with_data[:5]) # Last 1 is original job_data_dc    
    self.assertTrue(abs(sum_energy - slurm_monitor.total_energy) <= 10)

  def test_process_status_changes(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    CHUNK_UNIT = "month"
    CHUNK_SIZE = 20
    CURRENT_CONFIG = "CURRENT CONFIG"
    current_experiment_run_dc = exp_history.manager.get_experiment_run_dc_with_max_id()      
    exp_run = exp_history.process_status_changes(job_list=self.job_list, chunk_unit=CHUNK_UNIT, chunk_size=CHUNK_SIZE, current_config=CURRENT_CONFIG) # Generates new run    
    self.assertTrue(current_experiment_run_dc.run_id != exp_run.run_id)
    self.assertTrue(exp_run.chunk_unit == CHUNK_UNIT)
    self.assertTrue(exp_run.metadata == CURRENT_CONFIG)
    self.assertTrue(exp_run.total == len(self.job_list))
    current_experiment_run_dc = exp_history.manager.get_experiment_run_dc_with_max_id()      
    exp_run = exp_history.process_status_changes(job_list=self.job_list, chunk_unit=CHUNK_UNIT, chunk_size=CHUNK_SIZE, current_config=CURRENT_CONFIG) # Same run
    self.assertTrue(current_experiment_run_dc.run_id == exp_run.run_id)
    new_job_list = [
      job("a29z_20000101_fc2_1_POST", "2000-01-01 00:00:00", "POST", "FAILED", ""),
      job("a29z_20000101_fc1_1_CLEAN", "2000-01-01 00:00:00", "CLEAN", "FAILED", ""),
      job("a29z_20000101_fc3_1_POST", "2000-01-01 00:00:00", "POST", "RUNNING", ""),
      job("a29z_20000101_fc2_1_CLEAN", "2000-01-01 00:00:00", "CLEAN", "FAILED", ""),
      job("a29z_20000101_fc0_3_SIM", "2000-01-01 00:00:00", "SIM", "FAILED", ""),
      job("a29z_20000101_fc1_2_POST", "2000-01-01 00:00:00", "POST", "QUEUING", ""),
    ]
    current_experiment_run_dc = exp_history.manager.get_experiment_run_dc_with_max_id()
    exp_run = exp_history.process_status_changes(job_list=new_job_list, chunk_unit=CHUNK_UNIT, chunk_size=CHUNK_SIZE, current_config=CURRENT_CONFIG) # Generates new run
    self.assertTrue(current_experiment_run_dc.run_id != exp_run.run_id)
    self.assertTrue(exp_run.total == len(new_job_list))    
    self.assertTrue(exp_run.failed == 4)

  def test_write_submit_time(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    JOB_NAME = "a29z_20000101_fc2_1_SIM"
    NCPUS = 128
    PLATFORM_NAME = "marenostrum4"
    JOB_ID = 101
    inserted_job_data_dc = exp_history.write_submit_time(JOB_NAME, time.time(), "SUBMITTED", NCPUS, "00:30", "debug", "20000101", "fc2", "SIM", 1, PLATFORM_NAME, JOB_ID, "bsc_es", 1, "")
    self.assertTrue(inserted_job_data_dc.job_name == JOB_NAME)
    self.assertTrue(inserted_job_data_dc.ncpus == NCPUS)
    self.assertTrue(inserted_job_data_dc.children == "")
    self.assertTrue(inserted_job_data_dc.energy == 0)
    self.assertTrue(inserted_job_data_dc.platform == PLATFORM_NAME)
    self.assertTrue(inserted_job_data_dc.job_id == JOB_ID)
    self.assertTrue(inserted_job_data_dc.qos == "debug")


  def test_write_start_time(self):
    exp_history = ExperimentHistory("tt00")
    exp_history.initialize_database()
    JOB_NAME = "a29z_20000101_fc2_1_SIM"
    NCPUS = 128
    PLATFORM_NAME = "marenostrum4"
    JOB_ID = 101
    inserted_job_data_dc_submit = exp_history.write_submit_time(JOB_NAME, time.time(), "SUBMITTED", NCPUS, "00:30", "debug", "20000101", "fc2", "SIM", 1, PLATFORM_NAME, JOB_ID, "bsc_es", 1, "")
    inserted_job_data_dc = exp_history.write_start_time(JOB_NAME, time.time(), "RUNNING", NCPUS, "00:30", "debug", "20000101", "fc2", "SIM", 1, PLATFORM_NAME, JOB_ID, "bsc_es", 1, "")
    self.assertTrue(inserted_job_data_dc.job_name == JOB_NAME)
    self.assertTrue(inserted_job_data_dc.ncpus == NCPUS)
    self.assertTrue(inserted_job_data_dc.children == "")
    self.assertTrue(inserted_job_data_dc.energy == 0)
    self.assertTrue(inserted_job_data_dc.platform == PLATFORM_NAME)
    self.assertTrue(inserted_job_data_dc.job_id == JOB_ID)
    self.assertTrue(inserted_job_data_dc.status == "RUNNING")
    self.assertTrue(inserted_job_data_dc.qos == "debug")



class TestLogging(unittest.TestCase):

  def setUp(self):
    message = "No Message"
    try:
      raise Exception("Setup test exception")
    except Exception as e:
      message = traceback.format_exc()
    self.log = Logging("tt00")
    self.exp_message = "Exception message"
    self.trace_message = message

  def test_build_message(self):
    message = self.log.build_message(self.exp_message, self.trace_message)
    # print(message)
    self.assertIsNotNone(message)
    self.assertTrue(len(message) > 0)

  def test_log(self):    
    self.log.log(self.exp_message, self.trace_message)
    


if __name__ == '__main__':
  unittest.main()


from unittest import TestCase

import os
from pathlib import Path
import inspect
import tempfile
from mock import MagicMock
from mock import patch

from autosubmit.job.job import Job
from autosubmit.job.job_common import Status
from autosubmit.job.job_list import JobList
from autosubmit.job.job_list_persistence import JobListPersistenceDb
from autosubmit.job.job_packages import JobPackageSimple, JobPackageVertical
from autosubmitconfigparser.config.configcommon import AutosubmitConfig
from autosubmitconfigparser.config.yamlparser import YAMLParserFactory


class FakeBasicConfig:
    def __init__(self):
        pass
    def props(self):
        pr = {}
        for name in dir(self):
            value = getattr(self, name)
            if not name.startswith('__') and not inspect.ismethod(value) and not inspect.isfunction(value):
                pr[name] = value
        return pr
    DB_DIR = '/dummy/db/dir'
    DB_FILE = '/dummy/db/file'
    DB_PATH = '/dummy/db/path'
    LOCAL_ROOT_DIR = '/dummy/local/root/dir'
    LOCAL_TMP_DIR = '/dummy/local/temp/dir'
    LOCAL_PROJ_DIR = '/dummy/local/proj/dir'
    DEFAULT_PLATFORMS_CONF = ''
    DEFAULT_JOBS_CONF = ''

class TestJobPackage(TestCase):

    def setUpWrappers(self,options):
        # reset
        self.as_conf = None
        self.job_package_wrapper = None
        self.experiment_id = 'random-id'
        self._wrapper_factory = MagicMock()

        self.config = FakeBasicConfig
        self.config.read = MagicMock()


        with patch.object(Path, 'exists') as mock_exists:
            mock_exists.return_value = True
            self.as_conf = AutosubmitConfig(self.experiment_id, self.config, YAMLParserFactory())
        self.as_conf.experiment_data = dict()
        self.as_conf.experiment_data["JOBS"] = dict()
        self.as_conf.experiment_data["PLATFORMS"] = dict()
        self.as_conf.experiment_data["WRAPPERS"] = dict()
        self.temp_directory = tempfile.mkdtemp()
        self.job_list = JobList(self.experiment_id, self.config, YAMLParserFactory(),
                                JobListPersistenceDb(self.temp_directory, 'db'), self.as_conf)
        self.parser_mock = MagicMock(spec='SafeConfigParser')

        self.platform.max_waiting_jobs = 100
        self.platform.total_jobs = 100
        self.as_conf.experiment_data["WRAPPERS"]["WRAPPERS"] = options
        self._wrapper_factory.as_conf = self.as_conf
        self.jobs[0].wallclock = "00:00"
        self.jobs[0].threads = "1"
        self.jobs[0].tasks = "1"
        self.jobs[0].exclusive = True
        self.jobs[0].queue = "debug"
        self.jobs[0].partition = "debug"
        self.jobs[0].custom_directives = "dummy_directives"
        self.jobs[0].processors = "9"
        self.jobs[0]._processors = "9"
        self.jobs[0]._platform = self.platform
        self.jobs[0].retrials = 0
        self.jobs[1].wallclock = "00:00"
        self.jobs[1].threads = ""
        self.jobs[1].tasks = ""
        self.jobs[1].exclusive = True
        self.jobs[1].queue = "debug2"
        self.jobs[1].partition = "debug2"
        self.jobs[1].custom_directives = "dummy_directives2"
        self.jobs[1].processors = "9"
        self.jobs[1]._processors = "9"
        self.jobs[1]._platform = self.platform


        self.wrapper_type = options.get('TYPE', 'vertical')
        self.wrapper_policy = options.get('POLICY', 'flexible')
        self.wrapper_method = options.get('METHOD', 'ASThread')
        self.jobs_in_wrapper = options.get('JOBS_IN_WRAPPER', 'None')
        self.extensible_wallclock = options.get('EXTEND_WALLCLOCK', 0)
        self.job_package_wrapper = JobPackageVertical(self.jobs,configuration=self.as_conf,wrapper_info=[self.wrapper_type,self.wrapper_policy,self.wrapper_method,self.jobs_in_wrapper,self.extensible_wallclock])
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"] = dict()




    def setUp(self):
        self.platform = MagicMock()
        self.platform.queue = "debug"
        self.platform.partition = "debug"
        self.platform.serial_platform = self.platform
        self.platform.serial_queue = "debug-serial"
        self.platform.serial_partition = "debug-serial"
        self.jobs = [Job('dummy1', 0, Status.READY, 0),
                     Job('dummy2', 0, Status.READY, 0)]
        self.jobs[0]._platform = self.jobs[1]._platform = self.platform
        self.job_package = JobPackageSimple(self.jobs)
    def test_default_parameters(self):
        options = {
            'TYPE': "vertical",
            'JOBS_IN_WRAPPER': "None",
            'METHOD': "ASThread",
            'POLICY': "flexible",
            'EXTEND_WALLCLOCK': 0,
        }

        self.setUpWrappers(options)
        self.assertEqual(self.job_package_wrapper._wrapper_data["TYPE"], "vertical")
        self.assertEqual(self.job_package_wrapper._wrapper_data["JOBS_IN_WRAPPER"], "None")
        self.assertEqual(self.job_package_wrapper._wrapper_data["METHOD"], "ASThread")
        self.assertEqual(self.job_package_wrapper._wrapper_data["POLICY"], "flexible")
        self.assertEqual(self.job_package_wrapper._wrapper_data["EXTEND_WALLCLOCK"], 0)

        self.assertEqual(self.job_package_wrapper._wrapper_data["EXCLUSIVE"], True)
        self.assertEqual(self.job_package_wrapper._wrapper_data["INNER_RETRIALS"], 0)
        self.assertEqual(self.job_package_wrapper._wrapper_data["QUEUE"], "debug")
        self.assertEqual(self.job_package_wrapper._wrapper_data["PARTITION"], "debug")
        self.assertEqual(self.job_package_wrapper._wrapper_data["THREADS"], "1")
        self.assertEqual(self.job_package_wrapper._wrapper_data["TASKS"], "1")

        options_slurm = {
            'EXCLUSIVE': False,
            'QUEUE': "bsc32",
            'PARTITION': "bsc32",
            'THREADS': "30",
            'TASKS': "40",
            'INNER_RETRIALS': 30,
            'CUSTOM_DIRECTIVES': "['#SBATCH --mem=1000']"
        }
        self.setUpWrappers(options_slurm)
        self.assertEqual(self.job_package_wrapper._wrapper_data["EXCLUSIVE"], False)
        self.assertEqual(self.job_package_wrapper._wrapper_data["INNER_RETRIALS"], 30)
        self.assertEqual(self.job_package_wrapper._wrapper_data["QUEUE"], "bsc32")
        self.assertEqual(self.job_package_wrapper._wrapper_data["PARTITION"], "bsc32")
        self.assertEqual(self.job_package_wrapper._wrapper_data["THREADS"], "30")
        self.assertEqual(self.job_package_wrapper._wrapper_data["TASKS"], "40")
        self.assertEqual(self.job_package_wrapper._wrapper_data["CUSTOM_DIRECTIVES"], ['#SBATCH --mem=1000'])



    def test_job_package_default_init(self):
        with self.assertRaises(Exception):
            JobPackageSimple([])

    def test_job_package_different_platforms_init(self):
        self.jobs[0]._platform = MagicMock()
        self.jobs[1]._platform = MagicMock()
        with self.assertRaises(Exception):
            JobPackageSimple(self.jobs)

    def test_job_package_none_platforms_init(self):
        self.jobs[0]._platform = None
        self.jobs[1]._platform = None
        with self.assertRaises(Exception):
            JobPackageSimple(self.jobs)

    def test_job_package_length(self):
        self.assertEqual(2, len(self.job_package))

    def test_job_package_jobs_getter(self):
        self.assertEqual(self.jobs, self.job_package.jobs)

    def test_job_package_platform_getter(self):
        self.assertEqual(self.platform, self.job_package.platform)

    @patch("builtins.open",MagicMock())
    def test_job_package_submission(self):
        # arrange
        MagicMock().write = MagicMock()

        for job in self.jobs:
            job._tmp_path = MagicMock()
            job._get_paramiko_template = MagicMock("false","empty")

        self.job_package._create_scripts = MagicMock()
        self.job_package._send_files = MagicMock()
        self.job_package._do_submission = MagicMock()
        for job in self.jobs:
            job.update_parameters = MagicMock()
        # act
        self.job_package.submit('fake-config', 'fake-params')
        # assert
        for job in self.jobs:
            job.update_parameters.assert_called_once_with('fake-config', 'fake-params')
        self.job_package._create_scripts.is_called_once_with()
        self.job_package._send_files.is_called_once_with()
        self.job_package._do_submission.is_called_once_with()

    def test_wrapper_parameters(self):
        pass


from collections import namedtuple
from unittest import TestCase

from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import MagicMock

from autosubmit.platforms.slurmplatform import SlurmPlatform
from log.log import AutosubmitCritical, AutosubmitError


class TestSlurmPlatform(TestCase):

    Config = namedtuple('Config', ['LOCAL_ROOT_DIR', 'LOCAL_TMP_DIR', 'LOCAL_ASLOG_DIR'])

    def setUp(self):
        self.local_root_dir = TemporaryDirectory()
        self.config = {
            "LOCAL_ROOT_DIR" : self.local_root_dir.name,
            "LOCAL_TMP_DIR" : 'tmp',
            "LOCAL_ASLOG_DIR" : 'ASLOG_a000'
        }
        # We need to create the submission archive that AS expects to find in this location:
        p = Path(self.local_root_dir.name) / 'a000' / 'tmp' / 'ASLOG_a000'
        p.mkdir(parents=True)
        submit_platform_script = Path(p) / 'submit_local.sh'
        submit_platform_script.touch(exist_ok=True)

        self.platform = SlurmPlatform(expid='a000', name='local', config=self.config)

    def tearDown(self) -> None:
        self.local_root_dir.cleanup()

    def test_properties(self):
        props = {
            'name': 'foo',
            'host': 'localhost1',
            'user': 'sam',
            'project': 'proj1',
            'budget': 100,
            'reservation': 1,
            'exclusivity': True,
            'hyperthreading': True,
            'type': 'SuperSlurm',
            'scratch': '/scratch/1',
            'project_dir': '/proj1',
            'root_dir': '/root_1',
            'partition': 'inter',
            'queue': 'prio1'
        }
        for prop, value in props.items():
            setattr(self.platform, prop, value)
        for prop, value in props.items():
            self.assertEqual(value, getattr(self.platform, prop))

    def test_slurm_platform_submit_script_raises_autosubmit_critical_with_trace(self):
        package = MagicMock()
        package.jobs.return_value = []
        valid_packages_to_submit = [
            package
        ]

        ae = AutosubmitError(message='invalid partition', code=123, trace='ERR!')
        self.platform.submit_Script = MagicMock(side_effect=ae)

        # AS will handle the AutosubmitError above, but then raise an AutosubmitCritical.
        # This new error won't contain all the info from the upstream error.
        with self.assertRaises(AutosubmitCritical) as cm:
            self.platform.process_batch_ready_jobs(
                valid_packages_to_submit=valid_packages_to_submit,
                failed_packages=[]
            )

        # AS will handle the error and then later will raise another error message.
        # But the AutosubmitError object we created will have been correctly used
        # without raising any exceptions (such as AttributeError).
        assert cm.exception.message != ae.message



from unittest import TestCase
from unittest.mock import Mock,MagicMock, patch
from autosubmitconfigparser.config.configcommon import AutosubmitConfig
from autosubmitconfigparser.config.yamlparser import YAMLParserFactory
from autosubmit.autosubmit import Autosubmit
import autosubmit.platforms.pjmplatform
import pytest

from pathlib import Path
from autosubmit.platforms.platform import Platform
from autosubmit.platforms.pjmplatform import PJMPlatform
import autosubmit.platforms.headers.pjm_header
from tempfile import TemporaryDirectory
from datetime import datetime
from autosubmit.job.job import Job, Status
import inspect
class FakeBasicConfig:
    def props(self):
        pr = {}
        for name in dir(self):
            value = getattr(self, name)
            if not name.startswith('__') and not inspect.ismethod(value) and not inspect.isfunction(value):
                pr[name] = value
        return pr
    DB_DIR = '/dummy/db/dir'
    DB_FILE = '/dummy/db/file'
    DB_PATH = '/dummy/db/path'
    LOCAL_ROOT_DIR = '/dummy/local/root/dir'
    LOCAL_TMP_DIR = '/dummy/local/temp/dir'
    LOCAL_PROJ_DIR = '/dummy/local/proj/dir'
    LOCAL_ASLOG_DIR = '/dummy/local/aslog/dir'
    DEFAULT_PLATFORMS_CONF = ''
    DEFAULT_JOBS_CONF = ''
    @staticmethod
    def read():
        return
class TestPJM(TestCase):

    def setUp(self) -> None:
        self.exp_id = 'a000'
        self.as_conf = MagicMock()
        with patch.object(Path, 'exists') as mock_exists:
            mock_exists.return_value = True
            self.as_conf = AutosubmitConfig(self.exp_id, FakeBasicConfig, YAMLParserFactory())
        self.as_conf.experiment_data = dict()
        self.as_conf.experiment_data["DEFAULT"] = dict()
        self.as_conf.experiment_data["DEFAULT"]["HPCARCH"] = "ARM"
        yml_file = Path(__file__).resolve().parent / "files/fake-jobs.yml"
        factory = YAMLParserFactory()
        parser = factory.create_parser()
        parser.data = parser.load(yml_file)
        self.as_conf.experiment_data.update(parser.data)
        yml_file = Path(__file__).resolve().parent / "files/fake-platforms.yml"
        factory = YAMLParserFactory()
        parser = factory.create_parser()
        parser.data = parser.load(yml_file)
        self.as_conf.experiment_data.update(parser.data)
        self.setUp_pjm()


    @patch("builtins.open",MagicMock())
    def setUp_pjm(self):
        MagicMock().write = MagicMock()
        MagicMock().os.path.join = MagicMock()
        self.section = 'ARM'
        self.submitted_ok = "[INFO] PJM 0000 pjsub Job 167661 submitted."
        self.submitted_fail = "[ERR.] PJM 0057 pjsub node=32 is greater than the upper limit (24)."
        self.out= """JOB_ID     ST  REASON                         
167727     EXT COMPLETED               
167728     RNO -               
167729     RNE -               
167730     RUN -               
167732     ACC -               
167733     QUE -               
167734     RNA -               
167735     RNP -               
167736     HLD ASHOLD               
167737     ERR -               
167738     CCL -               
167739     RJT -  
"""
        self.completed_jobs = ["167727"]
        self.running_jobs = ["167728","167729","167730"]
        self.queued_jobs = ["167732","167733","167734","167735","167736"]
        self.failed_jobs = ["167737","167738","167739"]
        self.jobs_that_arent_listed = ["3442432423", "238472364782", "1728362138712"]
        self.completed_jobs_cmd = "167727"
        self.running_jobs_cmd = "167728+167729+167730"
        self.queued_jobs_cmd = "167732+167733+167734+167735+167736"
        self.failed_jobs_cmd = "167737+167738+167739"
        self.jobs_that_arent_listed_cmd = "3442432423+238472364782+1728362138712"
        self.submitter = Autosubmit._get_submitter(self.as_conf)
        self.submitter.load_platforms(self.as_conf)
        self.remote_platform = self.submitter.platforms[self.section]

    def test_parse_Alljobs_output(self):
        """Test parsing of all jobs output."""
        for job_id in self.completed_jobs:
            assert self.remote_platform.parse_Alljobs_output(self.out,job_id) in self.remote_platform.job_status["COMPLETED"]
        for job_id in self.failed_jobs:
            assert self.remote_platform.parse_Alljobs_output(self.out,job_id) in self.remote_platform.job_status["FAILED"]
        for job_id in self.queued_jobs:
            assert self.remote_platform.parse_Alljobs_output(self.out,job_id) in self.remote_platform.job_status["QUEUING"]
        for job_id in self.running_jobs:
            assert self.remote_platform.parse_Alljobs_output(self.out,job_id) in self.remote_platform.job_status["RUNNING"]
        for job_id in self.jobs_that_arent_listed:
            assert self.remote_platform.parse_Alljobs_output(self.out,job_id) == []

    def test_get_submitted_job_id(self):
        """Test parsing of submitted job id."""
        output = self.remote_platform.get_submitted_job_id(self.submitted_ok)
        assert output == [167661]
        output = self.remote_platform.get_submitted_job_id(self.submitted_fail)
        assert output == []
    def test_parse_queue_reason(self):
        """Test parsing of queue reason."""
        output = self.remote_platform.parse_queue_reason(self.out, self.completed_jobs[0])
        assert output == "COMPLETED"






#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

from log.log import AutosubmitCritical, Log
from time import sleep
from autosubmitconfigparser.config.basicconfig import BasicConfig
from autosubmitconfigparser.config.configcommon import AutosubmitConfig
from autosubmit.history.experiment_history import ExperimentHistory
from autosubmit.database.db_common import check_experiment_exists
import datetime
import sys
from typing import List

def handle_start_time(start_time):
  # type: (str) -> None
  """ Wait until the supplied time. """
  if start_time:
    Log.info("User provided starting time has been detected.")
    # current_time = time()
    datetime_now = datetime.datetime.now()
    target_date = parsed_time = None
    try:
      # Trying first parse H:M:S
      parsed_time = datetime.datetime.strptime(start_time, "%H:%M:%S")
      target_date = datetime.datetime(datetime_now.year, datetime_now.month,
                                      datetime_now.day, parsed_time.hour, parsed_time.minute, parsed_time.second)
    except Exception as e:
      try:
          # Trying second parse y-m-d H:M:S
          target_date = datetime.datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S")
      except Exception as e:
          target_date = None
          Log.critical(
              "The string input provided as the starting time of your experiment must have the format 'H:M:S' or 'yyyy-mm-dd H:M:S'. Your input was '{0}'.".format(start_time))
          return
    # Must be in the future
    if target_date < datetime.datetime.now():
      Log.critical("You must provide a valid date into the future. Your input was interpreted as '{0}', which is considered past.\nCurrent time {1}.".format(
          target_date.strftime("%Y-%m-%d %H:%M:%S"), datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
      return
    # Starting waiting sequence
    Log.info("Your experiment will start execution on {0}\n".format(target_date.strftime("%Y-%m-%d %H:%M:%S")))
    # Check time every second
    while datetime.datetime.now() < target_date:
      elapsed_time = target_date - datetime.datetime.now()
      sys.stdout.write("\r{0} until execution starts".format(elapsed_time))
      sys.stdout.flush()
      sleep(1)

def handle_start_after(start_after, expid, BasicConfig):
  # type: (str, str, BasicConfig) -> None
  """ Wait until the start_after experiment has finished."""
  if start_after:
    Log.info("User provided expid completion trigger has been detected.")
    # The user tries to be tricky
    if str(start_after) == str(expid):
        Log.info(
            "Hey! What do you think is going to happen? In theory, your experiment will run again after it has been completed. Good luck!")
    # Check if experiment exists. If False or None, it does not exist
    if not check_experiment_exists(start_after):
        return None
    # Historical Database: We use the historical database to retrieve the current progress data of the supplied expid (start_after)
    exp_history = ExperimentHistory(start_after, jobdata_dir_path=BasicConfig.JOBDATA_DIR, historiclog_dir_path=BasicConfig.HISTORICAL_LOG_DIR)
    if exp_history.is_header_ready() is False:
        Log.critical("Experiment {0} is running a database version which is not supported by the completion trigger function. An updated DB version is needed.".format(
            start_after))
        return
    Log.info("Autosubmit will start monitoring experiment {0}. When the number of completed jobs plus suspended jobs becomes equal to the total number of jobs of experiment {0}, experiment {1} will start. Querying every 60 seconds. Status format Completed/Queuing/Running/Suspended/Failed.".format(
        start_after, expid))
    while True:
        # Query current run
        current_run = exp_history.manager.get_experiment_run_dc_with_max_id()
        if current_run and current_run.finish > 0 and current_run.total > 0 and current_run.completed + current_run.suspended == current_run.total:
            break
        else:
            sys.stdout.write(
                "\rExperiment {0} ({1} total jobs) status {2}/{3}/{4}/{5}/{6}".format(start_after, current_run.total, current_run.completed, current_run.queuing, current_run.running, current_run.suspended, current_run.failed))
            sys.stdout.flush()
        # Update every 60 seconds
        sleep(60)

def get_allowed_members(run_members, as_conf):
  # type: (str, AutosubmitConfig) -> List  
  if run_members:
    allowed_members = run_members.split()
    rmember = [rmember for rmember in allowed_members if rmember not in as_conf.get_member_list()]
    if len(rmember) > 0:
      raise AutosubmitCritical("Some of the members ({0}) in the list of allowed members you supplied do not exist in the current list " +
            "of members specified in the conf files.\nCurrent list of members: {1}".format(str(rmember), str(as_conf.get_member_list())))
    if len(allowed_members) == 0:
      raise AutosubmitCritical("Not a valid -rom --run_only_members input: {0}".format(str(run_members)))
    return allowed_members
  return []


#!/usr/bin/env python3

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS
# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import collections
import autosubmit.history.utils as HUtils
import autosubmit.history.database_managers.database_models as Models
from datetime import datetime, timedelta
from json import dumps , loads

class JobData(object):
    """
    Robust representation of a row in the job_data table of the experiment history database.
    """

    def __init__(self, _id, counter=1, job_name="None", created=None, modified=None, submit=0, start=0, finish=0, 
                status="UNKNOWN", rowtype=0, ncpus=0, wallclock="00:00", qos="debug", energy=0, date="", section="", 
                member="", chunk=0, last=1, platform="NA", job_id=0, extra_data="", nnodes=0, run_id=None, MaxRSS=0.0, 
                AveRSS=0.0, out="", err="", rowstatus=Models.RowStatus.INITIAL, children="", platform_output=""):
      """
      """
      self._id = _id
      self.counter = counter
      self.job_name = job_name
      self.created = HUtils.get_current_datetime_if_none(created)
      self.modified = HUtils.get_current_datetime_if_none(modified)
      self._submit = int(submit)
      self._start = int(start)
      self._finish = int(finish)
      self.status = status
      self.rowtype = rowtype
      self.ncpus = ncpus
      self.wallclock = wallclock
      self.qos = qos if qos else "debug"
      self._energy = round(energy, 2) if energy else 0
      self.date = date if date else ""
      self.section = section if section else ""
      self.member = member if member else ""
      self.chunk = chunk if chunk else 0
      self.last = last
      self._platform = platform if platform and len(
          platform) > 0 else "NA"
      self.job_id = job_id if job_id else 0
      # TODO jobs_data wilmer did this part... need to check that is loading yaml.
      self.extra_data_parsed = {} # Fail fast
      try:
          if extra_data != "":
            self.extra_data_parsed = loads(extra_data)
      except Exception as exp:
          self.extra_data_parsed = {} # Fail fast
      self.extra_data = extra_data
      self.nnodes = nnodes
      self.run_id = run_id
      self.require_update = False
      # DB VERSION 15 attributes
      self.MaxRSS = MaxRSS
      self.AveRSS = AveRSS
      self.out = out
      self.err = err
      self.rowstatus = rowstatus      
      self.children = children # DB 17
      self.platform_output = platform_output # DB 17

    @classmethod
    def from_model(cls, row):
      """ Build JobData from JobDataRow. """
      job_data = cls(row.id, 
                      row.counter,
                      row.job_name,
                      row.created,
                      row.modified,
                      row.submit,
                      row.start,
                      row.finish,
                      row.status,
                      row.rowtype,
                      row.ncpus,
                      row.wallclock,
                      row.qos,
                      row.energy,
                      row.date,
                      row.section,
                      row.member,
                      row.chunk,
                      row.last,
                      row.platform,
                      row.job_id,
                      row.extra_data,
                      row.nnodes,
                      row.run_id,
                      row.MaxRSS,
                      row.AveRSS,
                      row.out,
                      row.err,
                      row.rowstatus,
                      row.children,
                      row.platform_output)
      return job_data

    @property
    def children_list(self):
        children_list = self.children.split(",") if self.children else []
        result = [str(job_name).strip() for job_name in children_list]
        return result

    @property
    def computational_weight(self):
        return round(float(self.running_time * self.ncpus),4)

    @property
    def submit(self):
        """
        Returns to submit time timestamp as an integer.
        """
        return int(self._submit)

    @property
    def start(self):
        """
        Returns the start time timestamp as an integer.
        """
        return int(self._start)

    @property
    def finish(self):
        """
        Returns the finish time timestamp as an integer.
        """
        return int(self._finish)

    @property
    def platform(self):
        """
        Returns the name of the platform, "NA" if no platform is set.
        """
        return self._platform

    @property
    def energy(self):
        """
        Returns the energy spent value (JOULES) as an integer.
        """
        return self._energy
    
    @property
    def wrapper_code(self):
        """ 
        Another name for rowtype
        """
        if self.rowtype > 2:        
            return self.rowtype
        else:
            return None

    @submit.setter
    def submit(self, submit):
        self._submit = int(submit)

    @start.setter
    def start(self, start):
        self._start = int(start)

    @finish.setter
    def finish(self, finish):
        self._finish = int(finish)

    @platform.setter
    def platform(self, platform):
        self._platform = platform if platform and len(platform) > 0 else ""

    @energy.setter
    def energy(self, energy):
        """
        Set the energy value. If it is different from the current energy value, an update flag will be activated.
        """
        if energy > 0:
            if energy != self._energy:
                # print("Updating energy to {0} from {1}.".format(
                #    energy, self._energy))
                self.require_update = True
            self._energy = round(energy, 2)

    @property
    def delta_queue_time(self):
        """
        Returns queuing time as a timedelta object.
        """
        return str(timedelta(seconds=self.queuing_time))

    @property
    def delta_running_time(self):
        """
        Returns running time as a timedelta object.
        """
        return str(timedelta(seconds=self.running_time))

    @property
    def submit_datetime(self):
        """
        Return to submit time as a datetime object, None if submit time equal 0.
        """
        if self.submit > 0:
            return datetime.fromtimestamp(self.submit)
        return None

    @property
    def start_datetime(self):
        """
        Return the start time as a datetime object, None if start time equal 0.
        """
        if self.start > 0:
            return datetime.fromtimestamp(self.start)
        return None

    @property
    def finish_datetime(self):
        """
        Return the finish time as a datetime object, None if start time equal 0.
        """
        if self.finish > 0:
            return datetime.fromtimestamp(self.finish)
        return None

    @property
    def submit_datetime_str(self):
        """
        Returns to submit datetime as a string with format %Y-%m-%d-%H:%M:%S
        """
        o_datetime = self.submit_datetime
        if o_datetime:
            return o_datetime.strftime(HUtils.DATETIME_FORMAT)
        else:
            return None
    @property
    def start_datetime_str(self):
        """
        Returns the start datetime as a string with format %Y-%m-%d-%H:%M:%S
        """
        o_datetime = self.start_datetime
        if o_datetime:
            return o_datetime.strftime(HUtils.DATETIME_FORMAT)
        else:
            return None
    @property
    def finish_datetime_str(self):
        """
        Returns the finish datetime as a string with format %Y-%m-%d-%H:%M:%S
        """
        o_datetime = self.finish_datetime
        if o_datetime:
            return o_datetime.strftime(HUtils.DATETIME_FORMAT)
        else:
            return None

    @property
    def running_time(self):
        """
        Calculates and returns the running time of the job, in seconds.

        :return: Running time in seconds.   
        :rtype: int
        """
        if self.status in ["RUNNING", "COMPLETED", "FAILED"]:
            return HUtils.calculate_run_time_in_seconds(self.start, self.finish)                        
        return 1

    @property
    def queuing_time(self):
        """
        Calculates and returns the queuing time of the job, in seconds.

        :return: Queueing time in seconds.   
        :rtype: int
        """
        if self.status in ["SUBMITTED", "QUEUING", "RUNNING", "COMPLETED", "HELD", "PREPARED", "FAILED", "SKIPPED"]:
            return HUtils.calculate_queue_time_in_seconds(self.submit, self.start)            
        return 0

    def get_hdata(self):
        """
        Get the job data as an ordered dict into a JSON object.  
        :return: Job data as an ordered dict into a JSON object.  
        :rtype: JSON object.
        """
        hdata = collections.OrderedDict()
        hdata["name"] = self.job_name
        hdata["date"] = self.date
        hdata["section"] = self.section
        hdata["member"] = self.member
        hdata["chunk"] = self.chunk
        hdata["submit"] = self.submit_datetime_str
        hdata["start"] = self.start_datetime_str
        hdata["finish"] = self.finish_datetime_str
        hdata["queue_time"] = self.delta_queue_time
        hdata["run_time"] = self.delta_running_time
        hdata["wallclock"] = self.wallclock
        hdata["ncpus"] = self.ncpus
        hdata["nnodes"] = self.nnodes
        hdata["energy"] = self.energy
        hdata["platform"] = self.platform
        hdata["MaxRSS"] = self.MaxRSS
        hdata["AveRSS"] = self.AveRSS
        return dumps(hdata)


#!/bin/env/python
import math
from autosubmit.job.job import Job
from datetime import datetime, timedelta
from autosubmit.job.job_common import Status
from typing import List, Tuple

from log.log import AutosubmitCritical

def filter_by_section(jobs, section):
  # type: (List[Job], str) -> List[Job]
  """ Filter jobs by provided section """
  if section and section != "Any":
    return [job for job in jobs if job.section == section]
  return jobs
  
def discard_ready_and_waiting(jobs):
  # type: (List[Job]) -> List[Job]
  if jobs and len(jobs) > 0:
    return [job for job in jobs if job.status not in [Status.READY, Status.WAITING]]
  return jobs

def filter_by_time_period(jobs, hours_span):
  # type: (List[Job], int) -> Tuple[List[Job], datetime, datetime]
  current_time = datetime.now().replace(second=0, microsecond=0)
  start_time = None
  if hours_span:
    if hours_span <= 0:
      raise AutosubmitCritical("{} is not a valid input for the statistics filter -fp.".format(hours_span))
    start_time = current_time - timedelta(hours=int(hours_span))
    return [job for job in jobs if job.check_started_after(start_time) or job.check_running_after(start_time)], start_time, current_time
  return jobs, start_time, current_time


def timedelta2hours(deltatime):
    # type: (timedelta) -> float
    return deltatime.days * 24 + deltatime.seconds / 3600.0

def parse_number_processors(processors_str):
  """ Defaults to 1 in case of error """
  if ':' in processors_str:
    components = processors_str.split(":")
    processors = int(sum(
        [math.ceil(float(x) / 36.0) * 36.0 for x in components]))
    return processors
  else:
    try:
      if processors_str == "":
        return 1
      else:
        processors = int(processors_str)
        return processors
    except Exception as e:
      return 1
  


from unittest import TestCase

import io
import sys
from contextlib import suppress, redirect_stdout
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch

from autosubmit.autosubmit import Autosubmit, AutosubmitCritical
from autosubmitconfigparser.config.basicconfig import BasicConfig


class TestJob(TestCase):

    def setUp(self):
        self.autosubmit = Autosubmit()
        # directories used when searching for logs to cat
        self.original_root_dir = BasicConfig.LOCAL_ROOT_DIR
        self.root_dir = TemporaryDirectory()
        BasicConfig.LOCAL_ROOT_DIR = self.root_dir.name
        self.exp_path = Path(self.root_dir.name, 'a000')
        self.tmp_dir = self.exp_path / BasicConfig.LOCAL_TMP_DIR
        self.aslogs_dir = self.tmp_dir / BasicConfig.LOCAL_ASLOG_DIR
        self.status_path = self.exp_path / 'status'
        self.aslogs_dir.mkdir(parents=True)
        self.status_path.mkdir()

    def tearDown(self) -> None:
        BasicConfig.LOCAL_ROOT_DIR = self.original_root_dir
        if self.root_dir is not None:
            self.root_dir.cleanup()

    def test_invalid_file(self):
        def _fn():
            self.autosubmit.cat_log(None, '8', None)  # type: ignore
        self.assertRaises(AutosubmitCritical, _fn)

    def test_invalid_mode(self):
        def _fn():
            self.autosubmit.cat_log(None, 'o', '8')  # type: ignore
        self.assertRaises(AutosubmitCritical, _fn)

    # -- workflow

    def test_is_workflow_invalid_file(self):
        def _fn():
            self.autosubmit.cat_log('a000', 'j', None)
        self.assertRaises(AutosubmitCritical, _fn)

    @patch('autosubmit.autosubmit.Log')
    def test_is_workflow_not_found(self, Log):
        self.autosubmit.cat_log('a000', 'o', 'c')
        assert Log.info.called
        assert Log.info.call_args[0][0] == 'No logs found.'

    def test_is_workflow_log_is_dir(self):
        log_file_actually_dir = Path(self.aslogs_dir, 'log_run.log')
        log_file_actually_dir.mkdir()
        def _fn():
            self.autosubmit.cat_log('a000', 'o', 'c')
        self.assertRaises(AutosubmitCritical, _fn)

    @patch('subprocess.Popen')
    def test_is_workflow_out_cat(self, popen):
        log_file = Path(self.aslogs_dir, 'log_run.log')
        with open(log_file, 'w') as f:
            f.write('as test')
            f.flush()
            self.autosubmit.cat_log('a000', file=None, mode='c')
            assert popen.called
            args = popen.call_args[0][0]
            assert args[0] == 'cat'
            assert args[1] == str(log_file)

    @patch('subprocess.Popen')
    def test_is_workflow_status_tail(self, popen):
        log_file = Path(self.status_path, 'a000_anything.txt')
        with open(log_file, 'w') as f:
            f.write('as test')
            f.flush()
            self.autosubmit.cat_log('a000', file='s', mode='t')
            assert popen.called
            args = popen.call_args[0][0]
            assert args[0] == 'tail'
            assert str(args[-1]) == str(log_file)

    # --- jobs

    @patch('autosubmit.autosubmit.Log')
    def test_is_jobs_not_found(self, Log):
        for file in ['j', 's', 'o']:
            self.autosubmit.cat_log('a000_INI', file=file, mode='c')
            assert Log.info.called
            assert Log.info.call_args[0][0] == 'No logs found.'

    def test_is_jobs_log_is_dir(self):
        log_file_actually_dir = Path(self.tmp_dir, 'LOG_a000/a000_INI.20000101.out')
        log_file_actually_dir.mkdir(parents=True)
        def _fn():
            self.autosubmit.cat_log('a000_INI', 'o', 'c')
        self.assertRaises(AutosubmitCritical, _fn)

    @patch('subprocess.Popen')
    def test_is_jobs_out_tail(self, popen):
        log_dir = self.tmp_dir / 'LOG_a000'
        log_dir.mkdir()
        log_file = log_dir / 'a000_INI.20200101.out'
        with open(log_file, 'w') as f:
            f.write('as test')
            f.flush()
            self.autosubmit.cat_log('a000_INI', file=None, mode='t')
            assert popen.called
            args = popen.call_args[0][0]
            assert args[0] == 'tail'
            assert str(args[-1]) == str(log_file)

    # --- command-line

    def test_command_line_help(self):
        args = ['autosubmit', 'cat-log', '--help']
        with patch.object(sys, 'argv', args) as _, io.StringIO() as buf, redirect_stdout(buf):
            with suppress(SystemExit):
                assert Autosubmit.parse_args()
            assert buf
            assert 'View workflow and job logs.' in buf.getvalue()





# Copyright 2017-2020 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

from autosubmit.database.db_manager import DbManager


class JobPackagePersistence(object):
    """
    Class that handles packages workflow.

    Create Packages Table, Wrappers Table.

    :param persistence_path: Path to the persistence folder pkl. \n
    :type persistence_path: String \n
    :param persistence_file: Name of the persistence pkl file. \n
    :type persistence_file: String
    """

    VERSION = 1
    JOB_PACKAGES_TABLE = 'job_package'
    WRAPPER_JOB_PACKAGES_TABLE = 'wrapper_job_package'
    TABLE_FIELDS = ['exp_id', 'package_name', 'job_name']

    def __init__(self, persistence_path, persistence_file):
        self.db_manager = DbManager(persistence_path, persistence_file, self.VERSION)
        self.db_manager.create_table(self.JOB_PACKAGES_TABLE, self.TABLE_FIELDS)
        self.db_manager.create_table(self.WRAPPER_JOB_PACKAGES_TABLE, self.TABLE_FIELDS)
    def load(self,wrapper=False):
        """
        Loads package of jobs from a database
        :param: wrapper: boolean
        :return: dictionary of jobs per package


        """
        if not wrapper:
            return self.db_manager.select_all(self.JOB_PACKAGES_TABLE)
        else:
            return self.db_manager.select_all(self.WRAPPER_JOB_PACKAGES_TABLE)
    def reset(self):
        """
        Loads package of jobs from a database

        """
        self.db_manager.drop_table(self.WRAPPER_JOB_PACKAGES_TABLE)
        self.db_manager.create_table(self.WRAPPER_JOB_PACKAGES_TABLE, self.TABLE_FIELDS)
    def save(self, package_name, jobs, exp_id,wrapper=False):
        """
        Persists a job list in a database
        :param package_name: str
        :param jobs: list of jobs
        :param exp_id: str
        :param wrapper: boolean


        """
        #self._reset_table()
        job_packages_data = []
        for job in jobs:
            job_packages_data += [(exp_id, package_name, job.name)]

        if  wrapper:
            self.db_manager.insertMany(self.WRAPPER_JOB_PACKAGES_TABLE, job_packages_data)
        else:
            self.db_manager.insertMany(self.JOB_PACKAGES_TABLE, job_packages_data)
            self.db_manager.insertMany(self.WRAPPER_JOB_PACKAGES_TABLE, job_packages_data)
    def reset_table(self,wrappers=False):
        """
        Drops and recreates the database
        """
        if wrappers:
            self.db_manager.drop_table(self.WRAPPER_JOB_PACKAGES_TABLE)
            self.db_manager.create_table(self.WRAPPER_JOB_PACKAGES_TABLE, self.TABLE_FIELDS)
        else:
            self.db_manager.drop_table(self.JOB_PACKAGES_TABLE)
            self.db_manager.create_table(self.JOB_PACKAGES_TABLE, self.TABLE_FIELDS)
            self.db_manager.drop_table(self.WRAPPER_JOB_PACKAGES_TABLE)
            self.db_manager.create_table(self.WRAPPER_JOB_PACKAGES_TABLE, self.TABLE_FIELDS)


import os
import pwd

from log.log import Log, AutosubmitCritical
from autosubmitconfigparser.config.basicconfig import BasicConfig
from typing import Tuple

def check_experiment_ownership(expid, basic_config, raise_error=False, logger=None):
  # [A-Za-z09]+ variable is not needed, LOG is global thus it will be read if available
  ## type: (str, BasicConfig, bool, Log) -> Tuple[bool, bool, str]
  my_user_ID = os.getuid()
  current_owner_ID = 0
  current_owner_name = "NA"
  try:
      current_owner_ID = os.stat(os.path.join(basic_config.LOCAL_ROOT_DIR, expid)).st_uid
      current_owner_name = pwd.getpwuid(os.stat(os.path.join(basic_config.LOCAL_ROOT_DIR, expid)).st_uid).pw_name
  except Exception as e:
      if logger:
        logger.info("Error while trying to get the experiment's owner information.")
  finally:
      if current_owner_ID <= 0 and logger:        
          logger.info("Current owner '{0}' of experiment {1} does not exist anymore.", current_owner_name, expid)
  is_owner = current_owner_ID == my_user_ID
  eadmin_user = os.popen('id -u eadmin').read().strip() # If eadmin no exists, it would be "" so INT() would fail.
  if eadmin_user != "":
      is_eadmin = my_user_ID == int(eadmin_user)
  else:
      is_eadmin = False
  if not is_owner and raise_error:
    raise AutosubmitCritical("You don't own the experiment {0}.".format(expid), 7012)     
  return is_owner, is_eadmin, current_owner_name


from .tests_commands import *
import os
import subprocess
import string

BIN_PATH = '../../bin'


def next_experiment_id(current_id):
    return base36encode(base36decode(current_id) + 1)


def check_cmd(command, path=BIN_PATH, verbose='AS_TEST_VERBOSE' in os.environ):
    try:
        output = subprocess.check_output(os.path.join(path, command), shell=True, stderr=subprocess.STDOUT)
        if verbose:
            print(output)

        if 'CRITICAL' in output or 'ERROR' in output:
            return False

        return True

    except subprocess.CalledProcessError as e:
        if verbose:
            print(e.output)
        return False


def copy_experiment_conf_files(db_path, src_path, experiment_id):
    check_cmd(get_copy_cmd(db_path, src_path, 'autosubmit', experiment_id), '')
    check_cmd(get_copy_cmd(db_path, src_path, 'expdef', experiment_id), '')
    check_cmd(get_copy_cmd(db_path, src_path, 'jobs', experiment_id), '')
    if os.path.exists(get_conf_file_path(src_path, 'platforms')):
        check_cmd(get_copy_cmd(db_path, src_path, 'platforms', experiment_id), '')
    else:
        check_cmd(get_default_copy_cmd(db_path, 'platforms', experiment_id), '')
    check_cmd(get_copy_cmd(db_path, src_path, 'proj', experiment_id), '')
    check_cmd(get_replace_exp_id(experiment_id) + os.path.join(db_path, experiment_id, 'conf', '*'), '', )
    check_cmd(get_replace_project_path(src_path) + os.path.join(db_path, experiment_id, 'conf', '*'), '', )


def clean_database(db_path):
    check_cmd('rm -rf ' + os.path.join(db_path, '*'), '', 'AS_TEST_VERBOSE' in os.environ)


def create_database():
    check_cmd(create_database_cmd())


def get_replace_exp_id(experiment_id):
    return "sed -i -- 's/EXPID-HERE/" + experiment_id + "/g' "


def get_replace_project_path(src_path):
    current_path = os.getcwd().replace('/', '\/')
    src_path = src_path.replace('/', '\/')
    return "sed -i -- 's/PROJECT-PATH-HERE/" + current_path + '\/' + src_path + '\/' + 'src' + "/g' "


def get_copy_cmd(db_path, src_path, filename, experiment_id):
    return 'cp ' + get_conf_file_path(src_path, filename) + ' ' + \
           get_conf_file_path(os.path.join(db_path, experiment_id), filename + '_' + experiment_id)


def get_default_copy_cmd(db_path, filename, experiment_id):
    return 'cp ' + os.path.join('default_conf', filename + '.conf') + ' ' + \
           get_conf_file_path(os.path.join(db_path, experiment_id), filename + '_' + experiment_id)


def get_conf_file_path(base_path, filename):
    return os.path.join(base_path, 'conf', filename + '.conf')


def base36encode(number, alphabet=string.digits + string.ascii_lowercase):
    """
    Convert positive integer to a base36 string.

    :param number: number to convert
    :type number: int
    :param alphabet: set of characters to use
    :type alphabet: str
    :return: number's base36 string value
    :rtype: str
    """
    if not isinstance(number, int):
        raise TypeError('number must be an integer')

    # Special case for zero
    if number == 0:
        return '0'

    base36 = ''

    sign = ''
    if number < 0:
        sign = '-'
        number = - number

    while number > 0:
        number, i = divmod(number, len(alphabet))
        # noinspection PyAugmentAssignment
        base36 = alphabet[i] + base36

    return sign + base36.rjust(4, '0')


def base36decode(number):
    """
    Converts a base36 string to a positive integer

    :param number: base36 string to convert
    :type number: str
    :return: number's integer value
    :rtype: int
    """
    return int(number, 36)



#!/usr/bin/env python3

# Copyright 2017-2020 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.
import locale
import os
from xml.dom.minidom import parseString
import subprocess


from autosubmit.platforms.paramiko_platform import ParamikoPlatform
from autosubmit.platforms.headers.local_header import LocalHeader

from autosubmitconfigparser.config.basicconfig import BasicConfig
from time import sleep
from log.log import Log, AutosubmitError, AutosubmitCritical

class LocalPlatform(ParamikoPlatform):
    """
    Class to manage jobs to localhost

    :param expid: experiment's identifier
    :type expid: str
    """

    def submit_Script(self, hold=False):
        pass

    def parse_Alljobs_output(self, output, job_id):
        pass

    def parse_queue_reason(self, output, job_id):
        pass

    def get_checkAlljobs_cmd(self, jobs_id):
        pass

    def __init__(self, expid, name, config):
        ParamikoPlatform.__init__(self, expid, name, config)
        self.cancel_cmd = None
        self.mkdir_cmd = None
        self.del_cmd = None
        self.get_cmd = None
        self.put_cmd = None
        self._checkhost_cmd = None
        self.type = 'local'
        self._header = LocalHeader()
        self.job_status = dict()
        self.job_status['COMPLETED'] = ['1']
        self.job_status['RUNNING'] = ['0']
        self.job_status['QUEUING'] = []
        self.job_status['FAILED'] = []
        self.update_cmds()

    def update_cmds(self):
        """
        Updates commands for platforms
        """
        self.root_dir = os.path.join(BasicConfig.LOCAL_ROOT_DIR, self.expid)
        self.remote_log_dir = os.path.join(self.root_dir, "tmp", 'LOG_' + self.expid)
        self.cancel_cmd = "kill -SIGINT"
        self._checkhost_cmd = "echo 1"
        self.put_cmd = "cp -p"
        self.get_cmd = "cp"
        self.del_cmd = "rm -f"
        self.mkdir_cmd = "mkdir -p " + self.remote_log_dir

    def get_checkhost_cmd(self):
        return self._checkhost_cmd

    def get_remote_log_dir(self):
        return self.remote_log_dir

    def get_mkdir_cmd(self):
        return self.mkdir_cmd

    def parse_job_output(self, output):
        return output[0]

    def get_submitted_job_id(self, output, x11 = False):
        return output

    def jobs_in_queue(self):
        dom = parseString('')
        jobs_xml = dom.getElementsByTagName("JB_job_number")
        return [int(element.firstChild.nodeValue) for element in jobs_xml]

    def get_submit_cmd(self, job_script, job, hold=False, export=""):
        wallclock = self.parse_time(job.wallclock)
        seconds = int(wallclock.days * 86400 + wallclock.seconds * 60)
        if export == "none" or export == "None" or export is None or export == "":
            export = ""
        else:
            export += " ; "
        return self.get_call(job_script, job, export=export,timeout=seconds)

    def get_checkjob_cmd(self, job_id):
        return self.get_pscall(job_id)

    def connect(self, reconnect=False):
        self.connected = True
    def test_connection(self):
        self.connected = True
    def restore_connection(self):
        self.connected = True

    def check_Alljobs(self, job_list, as_conf, retries=5):
        for job,prev_job_status in job_list:
            self.check_job(job)
    def send_command(self, command,ignore_log=False, x11 = False):
        lang = locale.getlocale()[1]
        if lang is None:
            lang = locale.getdefaultlocale()[1]
            if lang is None:
                lang = 'UTF-8'
        try:
            output = subprocess.check_output(command.encode(lang), shell=True)
        except subprocess.CalledProcessError as e:
            if not ignore_log:
                Log.error('Could not execute command {0} on {1}'.format(e.cmd, self.host))
            return False
        self._ssh_output = output.decode(lang)
        Log.debug("Command '{0}': {1}", command, self._ssh_output)

        return True

    def send_file(self, filename, check=True):
        self.check_remote_log_dir()
        self.delete_file(filename,del_cmd=True)
        command = '{0} {1} {2}'.format(self.put_cmd, os.path.join(self.tmp_path, filename),
                                       os.path.join(self.tmp_path, 'LOG_' + self.expid, filename))
        try:
            subprocess.check_call(command, shell=True)
        except subprocess.CalledProcessError:
            Log.error('Could not send file {0} to {1}'.format(os.path.join(self.tmp_path, filename),
                                                              os.path.join(self.tmp_path, 'LOG_' + self.expid,
                                                                           filename)))
            raise
        return True


    def get_file(self, filename, must_exist=True, relative_path='',ignore_log = False,wrapper_failed=False):
        local_path = os.path.join(self.tmp_path, relative_path)
        if not os.path.exists(local_path):
            os.makedirs(local_path)
        file_path = os.path.join(local_path, filename)
        if os.path.exists(file_path):
            os.remove(file_path)

        command = '{0} {1} {2}'.format(self.get_cmd, os.path.join(self.tmp_path, 'LOG_' + self.expid, filename),
                                       file_path)
        try:        
            subprocess.check_call(command, stdout=open(os.devnull, 'w'), stderr=open(os.devnull, 'w'), shell=True)                      
        except subprocess.CalledProcessError:
            if must_exist:
                raise Exception('File {0} does not exists'.format(filename))
            return False
        return True

    def check_remote_permissions(self):
        return True

    # Moves .err .out
    def check_file_exists(self, src,wrapper_failed=False):
        """
        Moves a file on the platform
        :param src: source name
        :type src: str
        :param: wrapper_failed: if True, the wrapper failed.
        :type wrapper_failed: bool

        """
        file_exist = False
        sleeptime = 5
        remote_path = os.path.join(self.get_files_path(), src)
        retries = 0
        max_retries = 3
        while not file_exist and retries < max_retries:
            try:
                file_exist = os.path.isfile(os.path.join(self.get_files_path(),src))
                if not file_exist:  # File doesn't exist, retry in sleep-time
                    Log.debug("{2} File does not exist.. waiting {0}s for a new retry (retries left: {1})", sleeptime,
                             max_retries - retries, remote_path)
                    if not wrapper_failed:
                        sleep(sleeptime)
                        sleeptime = sleeptime + 5
                        retries = retries + 1
                    else:
                        retries = 9999
            except BaseException as e:  # Unrecoverable error
                Log.printlog("File does not exist, logs {0} {1}".format(self.get_files_path(),src),6001)
                file_exist = False  # won't exist
                retries = 999  # no more retries
        return file_exist

    def delete_file(self, filename,del_cmd  = False):
        if del_cmd:
            command = '{0} {1}'.format(self.del_cmd, os.path.join(self.tmp_path,"LOG_"+self.expid, filename))
        else:
            command = '{0} {1}'.format(self.del_cmd, os.path.join(self.tmp_path,"LOG_"+self.expid, filename))
            command += ' ; {0} {1}'.format(self.del_cmd, os.path.join(self.tmp_path, filename))
        try:
            subprocess.check_call(command, shell=True)
        except subprocess.CalledProcessError:
            Log.debug('Could not remove file {0}'.format(os.path.join(self.tmp_path, filename)))
            return False
        return True
    def move_file(self, src, dest, must_exist=False):
        """
        Moves a file on the platform (includes .err and .out)
        :param src: source name
        :type src: str
        :param dest: destination name
        :param must_exist: ignore if file exist or not
        :type dest: str
        """
        path_root = ""
        try:
            path_root = self.get_files_path()
            os.rename(os.path.join(path_root, src),os.path.join(path_root, dest))
            return True
        except IOError as e:
            raise AutosubmitError('File {0} does not exists, something went wrong with the platform'.format(
                path_root), 6004, str(e))
            if must_exist:
                raise AutosubmitError("File {0} does not exists".format(
                    os.path.join(path_root,src)), 6004, str(e))
            else:
                Log.debug("File {0} doesn't exists ".format(path_root))
                return False
        except Exception as e:
            if str(e) in "Garbage":
                raise AutosubmitError('File {0} does not exists'.format(
                    os.path.join(self.get_files_path(), src)), 6004, str(e))
            if must_exist:
                raise AutosubmitError("File {0} does not exists".format(
                    os.path.join(self.get_files_path(), src)), 6004, str(e))
            else:
                Log.printlog("Log file couldn't be moved: {0}".format(
                    os.path.join(self.get_files_path(), src)), 5001)
                return False
    def get_ssh_output(self):
        return self._ssh_output
    def get_ssh_output_err(self):
        return self._ssh_output_err
    def get_logs_files(self, exp_id, remote_logs):
        """
        Overriding the parent's implementation.
        Do nothing because the log files are already in the local platform (redundancy).

        :param exp_id: experiment id
        :type exp_id: str
        :param remote_logs: names of the log files
        :type remote_logs: (str, str)
        """
        return



import logging
import os
import sys
from time import sleep
from datetime import datetime
from typing import Union


class AutosubmitError(Exception):
    """Exception raised for Autosubmit errors.

    Attributes:
        message (str): explanation of the error
        code (int): classified code
        trace (str): extra information about the error
    """

    def __init__(self, message="Unhandled Error", code=6000, trace: Union[None, str]=None):
        self.code = code
        self.message = message
        self.trace = trace

    @property
    def error_message(self) -> str:
        """
        Return the error message ready to be logged, with both trace
        (when present) and the message separated by a space. Or just
        the message if no trace is available.

        :return: ``trace`` and ``message`` separated by a space, or just the
                 ``message`` if no ``trace`` is available.
        :rtype: str
        """
        return self.message if not self.trace else f'{self.trace} {self.message}'

    def __str__(self):
        return " "


class AutosubmitCritical(Exception):
    """Exception raised for Autosubmit critical errors .
    Attributes:
        errorcode -- Classified code
        message -- explanation of the error
    """

    def __init__(self, message="Unhandled Error", code=7000, trace=None):
        self.code = code
        self.message = message
        self.trace = trace

    def __str__(self):
        return " "


class LogFormatter:

    """
    Class to format log output.

    :param to_file: If True, creates a LogFormatter for files; if False, for console
    :type to_file: bool
    """
    __module__ = __name__
    yellow = "\x1b[33;20m"
    RESULT = '\x1b[32m'
    ERROR = '\x1b[31m'
    CRITICAL = '\x1b[1m \x1b[31m'
    DEFAULT = '\x1b[0m\x1b[39m'
    ERROR = '\033[38;5;214m'
    WARNING = "\x1b[33;20m"

    def __init__(self, to_file=False):
        """
        Initializer for LogFormatter

        """
        self._file = to_file
        if self._file:
            self._formatter = logging.Formatter('%(asctime)s %(message)s')
        else:
            self._formatter = logging.Formatter('%(message)s')

    def format(self, record):
        """
        Format log output, adding labels if needed for log level. If logging to console, also manages font color.
        If logging to file adds timestamp

        :param record: log record to format
        :type record: LogRecord
        :return: formatted record
        :rtype: str
        """
        header = ''
        if record.levelno == Log.RESULT:
            if not self._file:
                header = LogFormatter.RESULT
        elif record.levelno == Log.WARNING:
            if not self._file:
                header = LogFormatter.WARNING
            header += '[WARNING] '
        elif record.levelno == Log.ERROR:
            if not self._file:
                header = LogFormatter.ERROR
            header += '[ERROR] '
        elif record.levelno == Log.CRITICAL:
            if not self._file:
                header = LogFormatter.CRITICAL
            header += '[CRITICAL] '
        msg = self._formatter.format(record)
        if header != '' and not self._file:
            msg += LogFormatter.DEFAULT
        return header + msg


class StatusFilter(logging.Filter):

    def filter(self, rec):
        return rec.levelno == Log.STATUS

class StatusFailedFilter(logging.Filter):
    def filter(self, rec):
        return rec.levelno == Log.STATUS_FAILED

class Log:
    """
    Static class to manage the log for the application. Messages will be sent to console and to file if it is
    configured. Levels can be set for each output independently. These levels are (from lower to higher priority):
    """

    date = ('{0:%Y%m%d_%H%M%S}_').format(datetime.now())

    def __init__(self):
        pass

    file_path = ""
    __module__ = __name__
    EVERYTHING = 0
    STATUS_FAILED = 500
    STATUS = 1000
    DEBUG = 2000
    WARNING = 3000
    INFO = 4000
    RESULT = 5000
    ERROR = 6000
    CRITICAL = 7000
    NO_LOG = CRITICAL + 1000
    logging.basicConfig()
    log_dict_debug = logging.Logger.manager.loggerDict
    if 'Autosubmit' in list(logging.Logger.manager.loggerDict.keys()):
        log = logging.getLogger('Autosubmit')
    else:
        log = logging.Logger('Autosubmit', EVERYTHING)
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(INFO)
    console_handler.setFormatter(LogFormatter(False))
    log.addHandler(console_handler)
    def init_variables(self,file_path=""):
        self.file_path = file_path

    @staticmethod
    def shutdown_logger():
        """
        Shutdown logger module to prevent race issues on delete
        """
        logging.shutdown()

    @staticmethod
    def get_logger(name="Autosubmit"):
        """
        Configure the file to store the log. If another file was specified earlier, new messages will only go to the
        new file.

        :param file_path: file to store the log
        :type file_path: str
        """
        logging.getLogger(name)

    @staticmethod
    def set_file(file_path, type='out', level="WARNING"):
        """
        Configure the file to store the log. If another file was specified earlier, new messages will only go to the
        new file.

        :param file_path: file to store the log
        :type file_path: str
        """
        levels = {}
        levels["STATUS_FAILED"] = 500
        levels["STATUS"] = 1000
        levels["DEBUG"] = 2000
        levels["WARNING"] = 3000
        levels["INFO"] = 4000
        levels["RESULT"] = 5000
        levels["ERROR"] = 6000
        levels["CRITICAL"] = 7000
        levels["NO_LOG"] = levels["CRITICAL"] + 1000

        level = levels.get(str(level).upper(),"DEBUG")

        max_retrials = 3
        retrials = 0
        timeout = 5

        while not os.path.exists(file_path) and retrials < max_retrials:
            sleep(timeout*retrials)
            try:
                directory, filename = os.path.split(file_path)
                if not os.path.exists(directory):
                    os.mkdir(directory)
                files = [f for f in os.listdir(directory) if os.path.isfile(
                    os.path.join(directory, f)) and f.endswith(filename)]
                if len(files) >= 10:
                    files.sort()
                    os.remove(os.path.join(directory, files[0]))
                file_path = os.path.join(
                    directory, Log.date + filename)
                if type == 'out':
                    file_handler = logging.FileHandler(file_path, 'w')
                    file_handler.setLevel(level)
                    file_handler.setFormatter(LogFormatter(True))
                    Log.log.addHandler(file_handler)
                elif type == 'err':
                    err_file_handler = logging.FileHandler(file_path, 'w')
                    err_file_handler.setLevel(Log.ERROR)
                    err_file_handler.setFormatter(LogFormatter(True))
                    Log.log.addHandler(err_file_handler)
                elif type == 'status':
                    custom_filter = StatusFilter()
                    file_path = os.path.join(directory, filename)
                    status_file_handler = logging.FileHandler(file_path, 'w')
                    status_file_handler.setLevel(Log.STATUS)
                    status_file_handler.setFormatter(LogFormatter(False))
                    status_file_handler.addFilter(custom_filter)
                    Log.log.addHandler(status_file_handler)
                elif type == 'status_failed':
                    custom_filter = StatusFailedFilter()
                    file_path = os.path.join(directory, filename)
                    status_file_handler = logging.FileHandler(file_path, 'w')
                    status_file_handler.setLevel(Log.STATUS_FAILED)
                    status_file_handler.setFormatter(LogFormatter(False))
                    status_file_handler.addFilter(custom_filter)
                    Log.log.addHandler(status_file_handler)
                os.chmod(file_path, 509)
            except: # retry again
                pass

    @staticmethod
    def reset_status_file(file_path,type='status', level=WARNING):
        """
        Configure the file to store the log. If another file was specified earlier, new messages will only go to the
        new file.

        :param file_path: file to store the log
        :type file_path: str
        """
        try:
            #test = Log.log.handlers
            if type == 'status':
                while len(Log.log.handlers) > 3:
                    Log.log.handlers.pop()
                custom_filter = StatusFilter()
                status_file_handler = logging.FileHandler(file_path, 'w')
                status_file_handler.setLevel(Log.STATUS)
                status_file_handler.setFormatter(LogFormatter(False))
                status_file_handler.addFilter(custom_filter)
                Log.log.addHandler(status_file_handler)
            elif type == 'status_failed':
                custom_filter = StatusFailedFilter()
                status_file_handler = logging.FileHandler(file_path, 'w')
                status_file_handler.setLevel(Log.STATUS_FAILED)
                status_file_handler.setFormatter(LogFormatter(False))
                status_file_handler.addFilter(custom_filter)
                Log.log.addHandler(status_file_handler)
        except:  # retry again
            pass
    @staticmethod
    def set_console_level(level):
        """
        Sets log level for logging to console. Every output of level equal or higher to parameter level will be
        printed on console

        :param level: new level for console
        :return: None
        """
        if type(level) is str:
            level = getattr(Log, level)
        Log.console_handler.level = level

    @staticmethod
    def set_error_level(level):
        """
        Sets log level for logging to console. Every output of level equal or higher to parameter level will be
        printed on console

        :param level: new level for console
        :return: None
        """
        if type(level) is str:
            level = getattr(Log, level)
        Log.error.level = level

    @staticmethod
    def debug(msg, *args):
        """
        Sends debug information to the log

        :param msg: message to show
        :param args: arguments for message formating (it will be done using format() method on str)
        """
        Log.log.log(Log.DEBUG, msg.format(*args))

    @staticmethod
    def info(msg, *args):
        """
        Sends information to the log

        :param msg: message to show
        :param args: arguments for message formatting (it will be done using format() method on str)
        """
        Log.log.log(Log.INFO, msg.format(*args))

    @staticmethod
    def result(msg, *args):
        """
        Sends results information to the log. It will be shown in green in the console.

        :param msg: message to show
        :param args: arguments for message formating (it will be done using format() method on str)
        """
        Log.log.log(Log.RESULT, msg.format(*args))

    @staticmethod
    def warning(msg, *args):
        """
        Sends program warnings to the log. It will be shown in yellow in the console.

        :param msg: message to show
        :param args: arguments for message formatting (it will be done using format() method on str)
        """
        Log.log.log(Log.WARNING, msg.format(*args))

    @staticmethod
    def error(msg, *args):
        """
        Sends errors to the log. It will be shown in red in the console.

        :param msg: message to show
        :param args: arguments for message formatting (it will be done using format() method on str)
        """
        Log.log.log(Log.ERROR, msg.format(*args))

    @staticmethod
    def critical(msg, *args):
        """
        Sends critical errors to the log. It will be shown in red in the console.

        :param msg: message to show
        :param args: arguments for message formatting (it will be done using format() method on str)
        """
        Log.log.log(Log.CRITICAL, msg.format(*args))

    @staticmethod
    def status(msg, *args):
        """
        Sends status of jobs to the log. It will be shown in white in the console.

        :param msg: message to show
        :param args: arguments for message formatting (it will be done using format() method on str)
        """
        Log.log.log(Log.STATUS, msg.format(*args))

    @staticmethod
    def status_failed(msg, *args):
        """
        Sends failed status of jobs to the log. It will be shown in white in the console.

        :param msg: message to show
        :param args: arguments for message formatting (it will be done using format() method on str)
        """
        Log.log.log(Log.STATUS_FAILED, msg.format(*args))

    @staticmethod
    def printlog(message="Generic message", code=4000):
        """Log management for Autosubmit messages .
        Attributes:
            errorcode -- Classified code
            message -- explanation
        """
        if 4000 <= code < 5000:
            Log.info("{0}", message)
        elif 5000 <= code < 6000:
            Log.result("{0}", message)
        elif 3000 <= code < 4000:
            Log.warning("{1}[eCode={0}]", code, message)
        elif 6000 <= code < 7000:
            Log.error("{1}[eCode={0}]", code, message)
        elif code <= 7000:
            Log.critical("{1}[eCode={0}]", code, message)
        else:
            Log.info("{0}", message)
